/*
 * Originally from Matthew Johnson's Dbus-java implementation.
 * 
   D-Bus Java Implementation
   Copyright (c) 2005-2006 Matthew Johnson

   This program is free software; you can redistribute it and/or modify it
   under the terms of either the GNU General Public License Version 2 or the
   Academic Free Licence Version 2.1.

   Full licence texts are included in the COPYING file with this program.
*/
package org.freedesktop.minidbus;

import static org.freedesktop.minidbus.Gettext._;
import static org.freedesktop.minidbus.PalmIpcLogger.*;

import java.io.IOException;
import java.util.LinkedList;
import java.util.regex.Pattern;

import org.freedesktop.minidbus.exceptions.*;

import com.palm.luna.message.RdxReport;

import cx.ath.matthew.debug.Debug;

/**
 * Handles a DBus connection.
 * @author Ed Wei
 */
public abstract class AbstractConnection {

    static final Pattern dollar_pattern = Pattern.compile("[$]");
    public static final boolean EXCEPTION_DEBUG;
    static final boolean FLOAT_SUPPORT;
    static final int MAX_ARRAY_LENGTH = 67108864;
    static final int MAX_NAME_LENGTH = 255;

    /**
     * Timeout in us on checking the BUS for incoming messages and sending outgoing messages
     */
    protected static final int TIMEOUT = 5000000; //100000;
    /** Initial size of the pending calls map */
    private static final int PENDING_MAP_INITIAL_SIZE = 10;
    static final String BUSNAME_REGEX = "^[-_a-zA-Z][-_a-zA-Z0-9]*(\\.[-_a-zA-Z][-_a-zA-Z0-9]*)*$";
    static final String CONNID_REGEX = "^:[0-9]*\\.[0-9]*$";
    static final String OBJECT_REGEX = "^/([-_a-zA-Z0-9]+(/[-_a-zA-Z0-9]+)*)?$";

    protected boolean isRunning;
    protected Transport transport;
    
    protected String busAddress;
    
    protected SendingThread sendingThread;
    protected ReceivingThread receivingThread;

    EfficientQueue outgoing;
    
    // map serial -> MethodCall
    protected EfficientMap pendingCalls;
    
    /* Mapping from serial(long) -> message receivers */
    protected EfficientMapGeneric<MessageCallback<?>> pendingCallbacks;
    
    LinkedList<Error> pendingErrors;
    
    MessageFilter fMessageFilters;

    static {
        FLOAT_SUPPORT = (null != System.getenv("DBUS_JAVA_FLOATS"));
        EXCEPTION_DEBUG = (null != System.getenv("DBUS_JAVA_EXCEPTION_DEBUG"));
    }

    protected AbstractConnection(String address) throws DBusException
    {
        pendingCalls = new EfficientMap(PENDING_MAP_INITIAL_SIZE);
        pendingCallbacks = new EfficientMapGeneric<MessageCallback<?>>(64);
        
        outgoing = new EfficientQueue(PENDING_MAP_INITIAL_SIZE);
        pendingErrors = new LinkedList<Error>();
        
        isRunning = true;
        busAddress = address;
    }

    protected class ReceivingThread extends Thread
    {
        public ReceivingThread()
        {
            setName("DBusReceiving");
        }

        public void run()
        {
            try {
                Message m = null;
                while (isRunning) {
                    m = null;

                    // read from the wire
                    try {
                        // this blocks on outgoing being non-empty or a message being available.
                        m = readIncoming();
                        if (m != null) {
                            DEBUG("RX: "+m);
                            synchronized (this) { notifyAll(); }

                            handleMessage(m);

                            m = null;
                        }
                    } catch (Exception e) { 
                        System.err.println("Receiving Thread Error: " + e);
                        e.printStackTrace();
                        RdxReport.ReportAppFatal(e, Thread.currentThread());
                        if (e instanceof FatalException) {
                            try {
                                // FIXME need to represent this as a low-level call
                                //handleMessage(new org.freedesktop.DBus.Local.Disconnected("/"));
                            } catch (Exception ee) {
                                if (EXCEPTION_DEBUG ) ee.printStackTrace();           
                            }
                            disconnect();
                        }
                    }

                }
                synchronized (this) { notifyAll(); }
            } catch (Throwable e) {
                System.err.println("Receiving Thread Error: " + e);
                e.printStackTrace();
                RdxReport.ReportAppFatal(e, Thread.currentThread());
            }
            
            System.err.println("DBUS: Receving thread is exiting.  isRunning is " + isRunning);
        }
    }
    
    private void handleMessage(final Message reply)
    {
        long serial = reply.getReplySerial();
        if (serial > 0)
        {
            MethodCall       mc = null;
            
            /* This is the send sync path */
            synchronized(pendingCalls) {
                if (pendingCalls.contains(serial))
                    mc = pendingCalls.remove(serial);
            }
            
            if (null != mc)
            {
                if (reply instanceof MethodReturn)
                {
                    MethodReturn mr = (MethodReturn)reply;
                    mc.setReply(reply);
                    mr.setCall(mc);
                    return;
                }
                else if (reply instanceof Error)
                {
                    mc.setReply((Error)reply);
                    return;
                }
            }
            
            MessageCallback<?> receiver = null;            
            synchronized(pendingCallbacks) {
                if (pendingCallbacks.contains(serial))
                {
                    receiver = pendingCallbacks.get(serial);
                    if (receiver.isSingleShot())
                    {
                    	pendingCallbacks.remove(serial);
                    }
                }
            }
            if (null != receiver)
            {
                receiver.callback(receiver.getContext(), reply);
            }
            return;
        }
        
        if (reply instanceof DBusSignal)
            handleMessage((DBusSignal) reply);
        else if (reply instanceof MethodCall)
            handleMessage((MethodCall) reply);
        else if (reply instanceof MethodReturn)
            handleMessage((MethodReturn) reply);
        else if (reply instanceof Error)
            handleMessage((Error) reply);
    }

    private void handleMessage(final Error err) {
        
        DEBUG("Handling incoming error: "+err);
        synchronized (pendingErrors) {
            pendingErrors.addLast(err);
        }
        
        if (null != fMessageFilters)
        {
            fMessageFilters.handleMessage(err);
        }
    }

    private void handleMessage(final MethodReturn mr) {
        DEBUG("Handling incoming method return: "+mr);
        
        if (null != fMessageFilters)
        {
            fMessageFilters.handleMessage(mr);
        }
    }

    private void handleMessage(final MethodCall m) {
        if (null != fMessageFilters)
        {
            fMessageFilters.handleMessage(m);
        }
    }

    private void handleMessage(final DBusSignal m) {
        if (null != fMessageFilters)
        {
            fMessageFilters.handleMessage(m);
        }
    }

    private class SendingThread extends Thread
    {
       public SendingThread()
       {
          setName("DBusSending");
       }
       public void run()
       {
          Message m = null;

          if (Debug.debug) Debug.print(Debug.INFO, "Monitoring outbound queue");
          // block on the outbound queue and send from it
          while (isRunning) {
             synchronized (outgoing) {
                if (Debug.debug) Debug.print(Debug.VERBOSE, "Blocking");
                while (outgoing.size() == 0 && isRunning) 
                   try { outgoing.wait(); } catch (InterruptedException Ie) {}
                if (Debug.debug) Debug.print(Debug.VERBOSE, "Notified");
                if (outgoing.size() > 0)
                   m = outgoing.remove();
             }
             if (null != m) 
             {
                DEBUG("TX: "+m);
                sendMessage(m);
             }
             m = null;
          } 

          if (Debug.debug) Debug.print(Debug.INFO, "Flushing outbound queue and quitting");
          // flush the outbound queue before disconnect.
          do {
             synchronized (outgoing) {
                if (!outgoing.isEmpty())
                   m = outgoing.remove(); 
                else m = null;
             }
             sendMessage(m);
          } while (null != m);

          // close the underlying streams
       }
    }
    
    public void queueOutgoing(Message m)
    {
       synchronized (outgoing) {
          outgoing.add(m); 
          if (Debug.debug) Debug.print(Debug.DEBUG, "Notifying outgoing thread");
          outgoing.notifyAll();
       }
    }
    
    public void queueOutgoing(Message m, MessageCallback<?> receiver)
    {
    	if (receiver != null)
    	{
	    	synchronized (pendingCallbacks) {
	    		pendingCallbacks.put(m.getSerial(), receiver);	
			}
    	}
        queueOutgoing(m);
    }
    
	public void callCancel(long serial) {
		
		synchronized (pendingCallbacks) {
			if (pendingCallbacks.contains(serial))
			{
				pendingCallbacks.remove(serial);
			}
		}
	}
    
    protected void sendMessage(Message m)
    {
       try {

          if (m instanceof DBusSignal) 
             ((DBusSignal) m).appendbody(this);

          if (m instanceof MethodCall) {

             if (0 == (m.getFlags() & Message.Flags.NO_REPLY_EXPECTED))
                if (null == pendingCalls) 
                   ((MethodCall) m).setReply(new Error("org.freedesktop.DBus.Local", "org.freedesktop.DBus.Local.Disconnected", 0, "s", new Object[] { _("Disconnected") }));
                else synchronized (pendingCalls) {
                   pendingCalls.put(m.getSerial(),(MethodCall) m);
                }
          }

          transport.mout.writeMessage(m);
          
       } catch (Exception e) {
          if (EXCEPTION_DEBUG && Debug.debug) Debug.print(Debug.ERR, e);
          if (m instanceof MethodCall && e instanceof DBusExecutionException) 
             try {
                ((MethodCall)m).setReply(new Error(m, e));
             } catch (DBusException DBe) {}
          else if (m instanceof MethodCall)
             try {
                Debug.print(Debug.INFO, "Setting reply to "+m+" as an error");
                ((MethodCall)m).setReply(new Error(m, new DBusExecutionException(_("Message Failed to Send: ")+e.getMessage())));
             } catch (DBusException DBe) {}
          else if (m instanceof MethodReturn)
             try {
                transport.mout.writeMessage(new Error(m, e));
             } catch(DBusException IOe) {
                 if (EXCEPTION_DEBUG && Debug.debug) Debug.print(Debug.ERR, e);            

             } catch(IOException IOe) {
                if (EXCEPTION_DEBUG && Debug.debug) Debug.print(Debug.ERR, IOe);            
             }
          if (e instanceof IOException) disconnect();
       }
    }
    private Message readIncoming() throws DBusException 
    {
       if (null == transport) throw new NotConnected(_("No transport present"));
       Message m = null;
       try {
          m = transport.min.readMessage();
       } catch (IOException IOe) {
          throw new FatalDBusException(IOe.getMessage());
       }
       return m;
    }
    
    public void setMessageFilter(MessageFilter filter)
    {
        fMessageFilters = filter;
    }
    
    protected void listen()
    {
       // start listening
       receivingThread = new ReceivingThread();
       receivingThread.start();
       sendingThread = new SendingThread();
       sendingThread.start();
    }
    
    /** 
     * Disconnect from the Bus.
     */
    public void disconnect()
    {
       if (Debug.debug) Debug.print(Debug.INFO, "Disconnecting Abstract Connection");

       // stop the main thread
       isRunning = false;

       // unblock the sending thread.
       synchronized (outgoing) {
          outgoing.notifyAll();
       }

       // disconnect from the transport layer
       try {
          transport.disconnect();
       } catch (IOException IOe) {
          if (EXCEPTION_DEBUG && Debug.debug) Debug.print(Debug.ERR, IOe);            
       }
    }

    public void finalize()
    {
       disconnect();
    }
}
