/*
 * Originally from Matthew Johnson's Dbus-java implementation.
 * 
   D-Bus Java Implementation
   Copyright (c) 2005-2006 Matthew Johnson

   This program is free software; you can redistribute it and/or modify it
   under the terms of either the GNU General Public License Version 2 or the
   Academic Free Licence Version 2.1.

   Full licence texts are included in the COPYING file with this program.
*/
package org.freedesktop.minidbus;

import static org.freedesktop.minidbus.Gettext._;

import java.io.IOException;
import java.text.ParseException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.Vector;

import org.freedesktop.minidbus.exceptions.*;
import org.freedesktop.minidbus.types.UInt32;

import cx.ath.matthew.debug.Debug;

/** Handles a connection to DBus.
 * <p>
 * This is a Singleton class, only 1 connection to the SYSTEM or SESSION busses can be made.
 * Repeated calls to getConnection will return the same reference.
 * </p>
 */
public class DBusConnection extends AbstractConnection {   
 
    /**
     * System Bus
     */
    public static final int SYSTEM = 0;
    /**
     * Session Bus
     */
    public static final int SESSION = 1;

    public static final String DEFAULT_SYSTEM_BUS_ADDRESS = "unix:path=/var/run/dbus/system_bus_socket";

    private List<String> busnames;

    private static final Map<Object,DBusConnection> conn = new HashMap<Object,DBusConnection>();
    private int _refcount = 0;
    private Object _reflock = new Object();

    /**
     * Connect to the BUS. If a connection already exists to the specified Bus, a reference to it is returned.
     * @param address The address of the bus to connect to
     * @throws DBusException  If there is a problem connecting to the Bus.
     */
    public static DBusConnection getConnection(String address) throws DBusException
    {
       synchronized (conn) {
          DBusConnection c = conn.get(address);
          if (null != c) {
             synchronized (c._reflock) { c._refcount++; }
             return c;
          }
          else {
             c = new DBusConnection(address);
             conn.put(address, c);
             return c;
          }
       }
    }
    /**
     * Connect to the BUS. If a connection already exists to the specified Bus, a reference to it is returned.
     * @param bustype The Bus to connect to.
     * @see #SYSTEM
     * @see #SESSION
     * @throws DBusException  If there is a problem connecting to the Bus.
     */
    public static DBusConnection getConnection(int bustype) throws DBusException
    {
       synchronized (conn) {
          String s = null;
          switch (bustype) {
             case SYSTEM:
                s = System.getenv("DBUS_SYSTEM_BUS_ADDRESS");
                if (null == s) s = DEFAULT_SYSTEM_BUS_ADDRESS;
                break;
             case SESSION:
                s = System.getenv("DBUS_SESSION_BUS_ADDRESS");
                if (null == s) throw new DBusException(_("Cannot Resolve Session Bus Address"));
                break;
             default:
                throw new DBusException(_("Invalid Bus Type: ")+bustype);
          }
          DBusConnection c = conn.get(s);
          if (Debug.debug) Debug.print(Debug.VERBOSE, "Getting bus connection for "+s+": "+c);
          if (null != c) {
             synchronized (c._reflock) { c._refcount++; }
             return c;
          }
          else {
             if (Debug.debug) Debug.print(Debug.DEBUG, "Creating new bus connection to: "+s);
             c = new DBusConnection(s);
             conn.put(s, c);
             return c;
          }
       }
    }
    @SuppressWarnings("unchecked")
    private DBusConnection(String address) throws DBusException
    {
       super(address);
       
       busnames = new Vector<String>();

       synchronized (_reflock) {
          _refcount = 1; 
       }
    
       try {
    	  System.out.println("Attempting to connect to " + busAddress);
          transport = new Transport(busAddress, AbstractConnection.TIMEOUT);
       } catch (IOException IOe) {
          if (EXCEPTION_DEBUG && Debug.debug) Debug.print(Debug.ERR, IOe);            
          throw new DBusException(_("Failed to connect to bus ")+IOe.getMessage());
       } catch (ParseException Pe) {
          if (EXCEPTION_DEBUG && Debug.debug) Debug.print(Debug.ERR, Pe);            
          throw new DBusException(_("Failed to connect to bus ")+Pe.getMessage());
       }

       // start listening for calls
       listen();

       // FIXME do the bus registration via low level calls.
       MethodCall hello = new MethodCall(DBus.SERVICE_BUS, DBus.PATH_DBUS, DBus.INTERFACE_DBUS,
               "Hello", (byte) 0, null);

       // FIXME validate return message.
       queueOutgoing(hello);
       
       Message reply = hello.getReply(25000);
       
       if (reply instanceof MethodReturn)
       {
           if (Debug.debug) Debug.print("Hello to bus successful.  Got reply " + reply);
       } else
       {
           throw new DBusException(_("Failed to say \"Hello\" to bus.  Error message:")
                   + reply);
       }

       // register disconnect handlers
//       DBusSigHandler h = new _sighandler();
//       addSigHandlerWithoutMatch(org.freedesktop.DBus.Local.Disconnected.class, h);
//       addSigHandlerWithoutMatch(org.freedesktop.DBus.NameAcquired.class, h);

       // register ourselves
//       _dbus = getRemoteObject("org.freedesktop.DBus", "/org/freedesktop/DBus", DBus.class);
//       try {
//          busnames.add(_dbus.Hello());
//       } catch (DBusExecutionException DBEe) {
//          if (EXCEPTION_DEBUG && Debug.debug) Debug.print(Debug.ERR, DBEe);
//          throw new DBusException(DBEe.getMessage());
//       }
    }
    
    /** 
     * Request a bus name.
     * Request the well known name that this should respond to on the Bus.
     * @param busname The name to respond to. MUST be in dot-notation like "org.freedesktop.local"
     * @throws DBusException If the register name failed, or our name already exists on the bus.
     *  or if busname is incorrectly formatted.
     */
    public void requestBusName(String busname) throws DBusException
    {
       if (!busname.matches(BUSNAME_REGEX)||busname.length() > MAX_NAME_LENGTH)
          throw new DBusException(_("Invalid bus name"));

       MethodCall m = new MethodCall(DBus.SERVICE_BUS, DBus.PATH_DBUS, DBus.INTERFACE_DBUS,
               "RequestName", (byte)0, "su", busname,
               new UInt32(DBus.DBUS_NAME_FLAG_DO_NOT_QUEUE));
       queueOutgoing(m);
       
       Message reply = m.getReply(5000);
       
       synchronized (this.busnames) {    

           // FIXME do the following with low level calls.
//          UInt32 rv;
//          try { 
//             rv = _dbus.RequestName(busname, 
//                   new UInt32(DBus.DBUS_NAME_FLAG_REPLACE_EXISTING |
//                      DBus.DBUS_NAME_FLAG_DO_NOT_QUEUE));
//          } catch (DBusExecutionException DBEe) {
//             if (EXCEPTION_DEBUG && Debug.debug) Debug.print(Debug.ERR, DBEe);
//             throw new DBusException(DBEe.getMessage());
//          }
//          switch (rv.intValue()) {
//             case DBus.DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER: break;
//             case DBus.DBUS_REQUEST_NAME_REPLY_IN_QUEUE: throw new DBusException(_("Failed to register bus name"));
//             case DBus.DBUS_REQUEST_NAME_REPLY_EXISTS: throw new DBusException(_("Failed to register bus name"));
//             case DBus.DBUS_REQUEST_NAME_REPLY_ALREADY_OWNER: break;
//             default: break;
//          }
          this.busnames.add(busname);
       }
    }
    /**
     * Returns the unique name of this connection.
     */
    public String getUniqueName()
    {
       return busnames.get(0);
    }
    /**
     * Returns all the names owned by this connection.
     */
    public String[] getNames()
    {
       Set<String> names = new TreeSet<String>();
       names.addAll(busnames);
       return names.toArray(new String[0]);
    }
    
    /** 
     * Disconnect from the Bus.
     * This only disconnects when the last reference to the bus has disconnect called on it
     * or has been destroyed.
     */
    public void disconnect()
    {
       synchronized (conn) {
          synchronized (_reflock) {
             if (0 == --_refcount) {
                if (Debug.debug) Debug.print(Debug.INFO, "Disconnecting DBusConnection");
                // Set all pending messages to have an error.
                try {
                   Error err = new Error(
                         "org.freedesktop.DBus.Local" , "org.freedesktop.DBus.Local.Disconnected", 0, "s", new Object[] { _("Disconnected") });
                   synchronized (pendingCalls) {
                      long[] set = pendingCalls.getKeys();
                      for (long l: set) if (-1 != l) {
                         MethodCall m = pendingCalls.remove(l);
                         if (null != m)
                            m.setReply(err);
                      }
                      pendingCalls = null; 
                   }
                   synchronized (pendingErrors) {
                      pendingErrors.add(err);
                   }
                } catch (DBusException DBe) {}

                conn.remove(busAddress);
                super.disconnect();
             }
          }
       }
    }
}
